/**
 * P4
 * http://snarfed.org/space/p4
 *
 * Copyright 2001-2004
 * Ken Ashcraft <kashcraft+p4@gmail.com>
 * Ryan Barrett <p4@ryanb.org>
 * Maulik Shah <p4@maulik.net>
 * Nathan Stoll <nathanstoll+p4@gmail.com>
 *
 * @file receiveq.cpp
 *
 * Implements the interface described in receiveq.h.  Also contains some
 * of the p4client methods.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * TODO: get rid of lots of asserts (ex. in plugin_loaded, plugin_stopped)
 */

#include "receiveq.h"
#include "msg_hello.h"
#include "msg_data.h"
#include "protocol.h"
#include "lka.h"
#include "p4client.h"
#include "p4error.h"
#include "settings.h"
#include "debug.h"

#include <algorithm>


/*******************************************************************
 * PUBLIC METHODS
 *******************************************************************/
creceiveq::creceiveq()
{
  pthread_mutex_init(&mtx, NULL);
}

creceiveq::~creceiveq()
{
  cdebug::instance() << acquire("receiveq_debug") << "dtor()" << release();
  pthread_mutex_destroy(&mtx);
}

uint creceiveq::wipe()
{
  cdebug::instance() << acquire("receiveq_debug") << "wipe()" << release();

  pthread_mutex_lock(&mtx);

  map<ushort, received_data>::iterator queues_it = queues.begin();

  for (; queues_it != queues.end(); queues_it++) {
    received_data &q = queues_it->second;

    delete q.first;

    // can't iterate through a queue, so delete and pop instead
    while (!q.second.empty()) {
      delete q.second.front();
      q.second.pop();
    }
    assert(q.second.empty());
  }

  queues.clear();
  pthread_mutex_unlock(&mtx);
  return kerr_success;
}

uint creceiveq::plugin_loaded(ushort plugin_id)
{
  cdebug::instance() << acquire("receiveq_debug")
                     << "plugin_loaded(): id = " << plugin_id
                     << release();

  pthread_mutex_lock(&mtx);

  // error check
  if (plugin_id == 0)         // plugin id 0 is for system messages
    return kerr_bad_arg;
  map<ushort, received_data>::const_iterator check = queues.find(plugin_id);
  if (check != queues.end())  // make sure we don't load more than once
    return kerr_plugin_already_loaded;

  received_data incoming;
  incoming.first = NULL;
  queues[plugin_id] = incoming;

  pthread_mutex_unlock(&mtx);

  // update local user's listing in lka
  clka::instance().update_local_user();
  
  pthread_mutex_unlock(&mtx);

  // tell everyone that we started a new plugin
  cmsg_hello hello;
  hello.broadcast();

  return kerr_success;
}

uint creceiveq::plugin_stopped(ushort plugin_id)
{
  cdebug::instance() << acquire("receiveq_debug")
                     << "plugin_stopped(): id = " << plugin_id
                     << release();

  pthread_mutex_lock(&mtx);

  if (plugin_id == 0)         // plugin id 0 is for system messages
    return kerr_bad_arg;
  map<ushort, received_data>::iterator entry = queues.find(plugin_id);
  if (entry == queues.end())  // make sure we don't delete more than once
    return kerr_plugin_not_loaded;

  queues.erase(entry);

  pthread_mutex_unlock(&mtx);

  // update local user's listing in lka
  clka::instance().update_local_user();

  // tell everyone that we stopped a plugin
  cmsg_hello hello;
  hello.broadcast();

  return kerr_success;
}

uint creceiveq::enqueue(cmsg_data *msg)
{
  assert(msg);

  cdebug::instance() << acquire("receiveq_debug")
                     << "enqueueing " << *msg
                     << release();

  assert(msg->byte_order == cmsg::ehost);
  ushort plugin_id = msg->hdr.plugin_id;
  
  pthread_mutex_lock(&mtx);

  map<ushort, received_data>::const_iterator check = queues.find(plugin_id);
  
  if (check == queues.end()) {  // we don't care about msgs for this plugin
    pthread_mutex_unlock(&mtx);
    return kerr_unknown_plugin;
  }

  queues[plugin_id].second.push(msg);
  
  pthread_mutex_unlock(&mtx);
  return kerr_success;
}

uint creceiveq::peek(ushort plugin_id, int &size) const
{
  cdebug::instance() << acquire("receiveq_peek_debug")
                     << "peek(): id = " << plugin_id
                     << release();

  pthread_mutex_lock(&mtx);

  map<ushort, received_data>::const_iterator plugin = queues.find(plugin_id);
  assert(plugin != queues.end());  // make sure this plugin has loaded

  if (!plugin->second.second.empty()) {
    cmsg_data *next_msg = plugin->second.second.front();
    size = next_msg->hdr.data_len;
  } else
    size = 0;

  pthread_mutex_unlock(&mtx);

  return kerr_success;
}

uint creceiveq::dequeue(ushort plugin_id, char *user, char *data, int size,
            int &size_written, bool &encrypted)
{
  cdebug::instance() << acquire("receiveq_debug")
                     << "dequeue(): id = " << plugin_id << " buffer of size "
                     << size << release();

  if (!user || !data)
    return kerr_bad_arg;

  pthread_mutex_lock(&mtx);

  map<ushort, received_data>::const_iterator check = queues.find(plugin_id);
  assert(check != queues.end());  // make sure this plugin has loaded

  size_written = 0;
  
  if (!queues[plugin_id].second.empty()) {
    cmsg_data *next_msg = queues[plugin_id].second.front();
    cdebug::instance() << acquire("receiveq_debug")
                       << "dequeued msg " << *next_msg
                       << release();
    assert(next_msg->byte_order == cmsg::ehost);

    queues[plugin_id].second.pop();       // remove it...
    if (queues[plugin_id].first)
      delete queues[plugin_id].first;     // ...get rid of the old one...
    queues[plugin_id].first = next_msg;   // ...and hang on to the new one

    size_written = static_cast<int>(next_msg->hdr.data_len);

    if (size_written > size) {
      // the client provided buffer isn't big enough
      cdebug::instance() << acquire("receiveq_debug") 
                         << "error, datalen " << size_written << " is too big "
                         << next_msg->hdr.data_len
                         << release();

      size_written = 0;
      pthread_mutex_unlock(&mtx);
      return kerr_buffer_too_small;
    }

    // copy data into client buffer
    memcpy(data, next_msg->data, size_written);
    encrypted = next_msg->hdr.is_flag_set(kp4_encrypted);
    strcpy(user, next_msg->hdr.username);
  }

  pthread_mutex_unlock(&mtx);
  return ((size_written == 0) ? kerr_no_msg : kerr_success);
}

uint creceiveq::authenticate(ushort plugin_id, bool &authenticated) const
{
  cdebug::instance() << acquire("receiveq_debug")
                     << "authenticate()"
                     << release();
  // lock
  authenticated = false;
  // unlock
  return kerr_success;
}

ushort creceiveq::num_plugins() const
{
  cdebug::instance() << acquire("receiveq_debug")
                     << "num_plugins()"
                     << release();

  pthread_mutex_lock(&mtx);
  ushort num = queues.size();
  pthread_mutex_unlock(&mtx);

  return num;
}

uint creceiveq::get_plugins(ushort num_plugins, plugin_info *plugins, 
                            ushort &num_written) const
{
  cdebug::instance() << acquire("receiveq_debug")
                     << "get_plugins()"
                     << release();

  pthread_mutex_lock(&mtx);

  int i;
  map<ushort, received_data>::const_iterator it = queues.begin();
  for (i = 0; it != queues.end() && i < num_plugins; it++, i++)
    plugins[i].id = it->first;

  num_written = i;
  
  pthread_mutex_unlock(&mtx);
  return kerr_success;
}



#ifdef RECEIVEQ_MAIN
#include "msg_factory.h"

int main(int argc, char **argv)
{
  char *data = "I really like programming.  Programming is great!";
  p4_header hdr;
  hdr.set_type(kp4_data);
  hdr.plugin_id = 5;
  hdr.data_len = strlen(data) + 1;
  strcpy(hdr.username, "Ken");

  cmsg_data *msg = (cmsg_data*)cmsg_factory::instance().new_msg(hdr);
  
  msg->data = data;
  cout << "testing receiveq" << endl;

  uint ret = creceiveq::instance().plugin_loaded(5);
  assert(ret == kerr_success);

  ret = creceiveq::instance().enqueue(msg);
  assert(ret == kerr_success);
}

#endif
